import numpy as np
from charm_lensing.build_lens_system import build_lens_system
from charm_lensing.nifty_connect import update_position

import jax
jax.config.update("jax_platform_name", "cpu")

TEST_FOLDER = './data/charm_lensing_test'


def load_position():
    from os.path import join
    position = np.load(join(TEST_FOLDER, 'position.npy'), allow_pickle=True)
    position = position.item()
    return update_position(position, key_missing_update=True)


def load_field(field_name):
    from os.path import join
    field = np.load(join(TEST_FOLDER, f'{field_name}.npy'))
    return field


position = load_position()
lens_system = build_lens_system(
    dict(
        spaces=dict(
            source_space=dict(
                Npix=[64, 64],
                distance=0.04,
                padding_ratio=1.0
            ),
            lens_space=dict(
                Npix=[128, 128],
                distance=0.1,
                padding_ratio=2.0
            )
        ),

        model=dict(
            source=dict(
                light=dict(
                    mean=dict(
                        gaussian=dict(
                            center=dict(
                                distribution='normal',
                                mean=0.0,
                                sigma=0.2),
                            covariance=dict(
                                distribution='log_normal',
                                mean=0.2,
                                sigma=7e-2),
                            off_diagonal=dict(
                                distribution='normal',
                                mean=0.0,
                                sigma=10.0))
                    ),
                    perturbations=dict(
                        mattern=dict(
                            amplitude=dict(
                                offset_mean=0.0,
                                offset_std=[0.5e+0, 1e-1]),
                            fluctuations=dict(
                                non_parametric_kind='amplitude',
                                renormalize_amplitude=False,
                                scale=[1.5, 0.1],
                                cutoff=[0.22, 0.02],
                                loglogslope=[-4.0, 0.5])))
                )),

            lens=dict(
                poisson_padding=2,
                convergence=dict(
                    mean=dict(
                        piemd=dict(
                            b=dict(
                                distribution='log_normal',
                                mean=1.0,
                                sigma=0.2),
                            rs=dict(
                                distribution='log_normal',
                                mean=0.3,
                                sigma=0.2),
                            center=dict(
                                distribution='normal',
                                mean=0.0,
                                sigma=0.2),
                            q=dict(
                                distribution='uniform',
                                min=1.0,
                                max=2.4),
                            theta=dict(
                                distribution='normal',
                                mean=0.0,
                                sigma=10.0))),
                    perturbations=dict(
                        correlated_field=dict(
                            amplitude=dict(
                                offset_mean=0.0,
                                offset_std=[0.1e-5, 1e-6]),
                            fluctuations=dict(
                                fluctuations=[0.1, 0.05],
                                loglogavgslope=[-5.0, 0.5],
                                flexibility=[0.5, 1.0],
                                asperity=None))),
                    shear=dict(
                        shear=dict(
                            strength=dict(
                                distribution='normal',
                                mean=0.0,
                                sigma=0.07),
                            theta=dict(
                                distribution='normal',
                                mean=0.0,
                                sigma=10.0),
                            center=dict(
                                distribution=None,
                                mean=0.,
                                sigma=0.)
                        ))
                ),

                light=dict(
                    mean=dict(
                        sersic=dict(
                            ie=dict(
                                distribution='log_normal',
                                mean=0.001,
                                sigma=0.0001),
                            re=dict(
                                distribution='log_normal',
                                mean=1.5,
                                sigma=0.25),
                            n=dict(
                                distribution='uniform',
                                min=0.36,
                                max=10.0),
                            center=dict(
                                distribution='normal',
                                mean=0.0,
                                sigma=0.01),
                            q=dict(
                                distribution='uniform',
                                min=0.5,
                                max=10.5),
                            theta=dict(
                                distribution='normal',
                                mean=0.0,
                                sigma=10.0)))
                ),

            )
        )
    ))


def test_kappa():
    kappa_model = lens_system.lens_plane_model.convergence_model.model
    kappa = kappa_model(position)
    kappa_test = load_field('kappa')
    assert np.allclose(kappa, kappa_test)


def test_alpha():
    alpha_model = lens_system.lens_plane_model.convergence_model.deflection()
    alpha = alpha_model(position)
    alpha_test = load_field('alpha')
    assert np.allclose(alpha, alpha_test)


def test_lens_light():
    lens_light_model = lens_system.lens_plane_model.light_model.model
    lens_light = lens_light_model(position)
    lens_light_test = load_field('lens_light')
    assert np.allclose(lens_light, lens_light_test)


def test_source_light():
    source_light_model = lens_system.source_plane_model.light_model.model
    source_light = source_light_model(position)
    source_light_test = load_field('source_light')
    assert np.allclose(source_light, source_light_test)


def test_lensed_light():
    lensed_light_model = lens_system.get_forward_model_full()
    lensed_light = lensed_light_model(position)['sky']
    lensed_light_test = load_field('lensed_light')
    assert np.allclose(lensed_light, lensed_light_test)


def test_lensed_light_parametric():
    lensed_light_model = lens_system.get_forward_model_parametric()
    lensed_light = lensed_light_model(position)['sky']
    lensed_light_test = load_field('lensed_light_parametric')
    assert np.allclose(lensed_light, lensed_light_test)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from dataclasses import dataclass

from .spaces import Space
from .models import Model, build_hybrid_model


@ dataclass
class SpaceModel:
    space: Space
    light_model: Model | None = None
    convergence_model: Model | None = None


def build_space_model(
        space: Space,
        space_cfg: dict,
        light_extend: bool = False
) -> SpaceModel:

    convergence = build_hybrid_model(
        space.extend(),
        'convergence',
        space_cfg.get('convergence', None),
    )

    light = build_hybrid_model(
        space.extend() if light_extend else space,
        'light',
        space_cfg.get('light', None),
    )

    return SpaceModel(space, light, convergence)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from jax.scipy.signal import convolve2d, fftconvolve
from functools import partial

from typing import Callable
from numpy.typing import ArrayLike


def PsfOperator_dir(field, kernel):
    '''Creates a Psf-operator: convolution of field by kernel'''
    return convolve2d(field, kernel, mode='same')


def PsfOperator_fft(field, kernel):
    '''Creates a Psf-operator: convolution of field by kernel'''
    return fftconvolve(field, kernel, mode='same')


def instantiate_psf(psf: ArrayLike | None) -> Callable:
    if psf is None:
        return lambda x: x

    return partial(PsfOperator_fft, kernel=psf)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import nifty8.re as jft
from numpy.typing import ArrayLike


def build_gaussian_likelihood(
        data: ArrayLike,
        std: ArrayLike
):
    if not isinstance(std, float):
        assert data.shape == std.shape

    var = std**2

    return jft.Gaussian(
        data=data,
        noise_cov_inv=lambda x: x/var,
        noise_std_inv=lambda x: x/std
    )


def model_wrap(model, target_domain=None):
    if target_domain is None:
        def wrapper(x):
            out = model(x)
            for x, val in x.items():
                out[x] = val
            return out
    else:
        def wrapper(x):
            out = model(x)
            for key in target_domain.keys():
                out[key] = x[key]
            return out
    return wrapper


def connect_likelihood_to_model(
    likelihood: jft.Likelihood,
    model: jft.Model
) -> jft.Likelihood:

    ldom = likelihood.domain.tree
    tdom = {t: ldom[t] for t in ldom.keys() if t not in model.target.keys()}
    mdom = {t: model.domain[t] for t in model.domain.keys()}
    mdom.update(tdom)

    model_wrapper = model_wrap(model, tdom)

    model = jft.Model(
        lambda x: jft.Vector(model_wrapper(x)),
        domain=jft.Vector(mdom)
    )

    return likelihood.amend(model, domain=model.domain)



import pytest
import numpy as np
from charm_lensing.spaces import (
    coords, get_xycoords, get_klcoords, get_centered_slice,
    Space
)
from dataclasses import asdict
from nifty8 import RGSpace


def test_coords():
    # Test case 1
    shape = 5
    distance = 1
    expected_output = np.array([-2, -1, 0, 1, 2])
    assert np.array_equal(coords(shape, distance), expected_output)

    # Test case 2
    shape = 6
    distance = 0.5
    expected_output = np.array([-1.25, -0.75, -0.25, 0.25, 0.75,  1.25])
    assert np.allclose(coords(shape, distance), expected_output)


def test_get_xycoords():
    # Test case 1
    shape = (5, 5)
    distances = (1, 1)
    expected_output = np.array([
        [[-2, -1,  0,  1,  2],
         [-2, -1,  0,  1,  2],
         [-2, -1,  0,  1,  2],
         [-2, -1,  0,  1,  2],
         [-2, -1,  0,  1,  2]],

        [[-2, -2, -2, -2, -2],
         [-1, -1, -1, -1, -1],
         [0,  0,  0,  0,  0],
         [1,  1,  1,  1,  1],
         [2,  2,  2,  2,  2]]
    ])
    assert np.array_equal(get_xycoords(shape, distances), expected_output)

    # Test case 2
    shape = (3, 4)
    distances = (0.5, 1)
    expected_output = np.array([
        [[-0.5,  0.,  0.5],
         [-0.5,  0.,  0.5],
         [-0.5,  0.,  0.5],
         [-0.5,  0.,  0.5]],

        [[-1.5, -1.5, -1.5],
         [-0.5, -0.5, -0.5],
         [0.5,  0.5,  0.5],
         [1.5,  1.5,  1.5]]
    ])
    assert np.allclose(get_xycoords(shape, distances), expected_output)

    # Test case 3
    shape = (3, 4)
    distances = (1, 0.5)
    assert get_xycoords(shape, distances).shape == (2, 4, 3)


def test_get_klcoords():
    # Test case 1
    shape = (5, 5)
    distances = (1, 1)
    expected_output = np.array([
        [[0.,  0.2,  0.4, -0.4, -0.2],
         [0.,  0.2,  0.4, -0.4, -0.2],
         [0.,  0.2,  0.4, -0.4, -0.2],
         [0.,  0.2,  0.4, -0.4, -0.2],
         [0.,  0.2,  0.4, -0.4, -0.2]],

        [[0.,  0.,  0.,  0.,  0.],
         [0.2,  0.2,  0.2,  0.2,  0.2],
         [0.4,  0.4,  0.4,  0.4,  0.4],
         [-0.4, -0.4, -0.4, -0.4, -0.4],
         [-0.2, -0.2, -0.2, -0.2, -0.2]]])
    assert np.allclose(get_klcoords(shape, distances), expected_output)

    # Test case 2
    shape = (3, 4)
    distances = (0.5, 1)
    expected_output = np.array([
        [[0.,  0.25, -0.5, -0.25],
         [0.,  0.25, -0.5, -0.25],
         [0.,  0.25, -0.5, -0.25]],

        [[0.,  0.,  0.,  0.],
         [0.66666667,  0.66666667, 0.66666667,  0.66666667],
         [-0.66666667, -0.66666667, -0.66666667, -0.66666667]]
    ])
    assert np.allclose(get_klcoords(shape, distances), expected_output)

    # Test case 3
    shape = (2, 3, 4)
    distances = (2, 1, 0.5)
    with pytest.raises(AssertionError):
        get_klcoords(shape, distances)


def test_get_centered_slice():
    # Test case 1
    extended_shape = (9, 9)
    original_shape = (5, 5)
    expected_output = (slice(2, 7), slice(2, 7))
    assert get_centered_slice(
        extended_shape, original_shape) == expected_output

    # Test case 2
    extended_shape = (11, 12)
    original_shape = (3, 4)
    expected_output = (slice(4, 7), slice(4, 8))
    assert get_centered_slice(
        extended_shape, original_shape) == expected_output

    # Test case 3
    extended_shape = (7, 7, 7)
    original_shape = (3, 3, 3)
    expected_output = (slice(2, 5), slice(2, 5), slice(2, 5))
    assert get_centered_slice(
        extended_shape, original_shape) == expected_output

    # Test case 4
    extended_shape = (10, 10)
    original_shape = (5, 5)
    expected_output = (slice(None), slice(3, 8), slice(3, 8))
    assert get_centered_slice(
        extended_shape, original_shape, coordinates=True) == expected_output


def test_Space():
    # Test case 1
    shape = (5, 5)
    distances = (1, 1)
    extend_factor = 2
    space = Space(shape, distances, extend_factor=extend_factor)

    # Test that the shape and distances are correctly stored in the instance
    assert space.shape == shape
    assert space.distances == distances

    # Test that the coordinates are correctly generated
    expected_coords = get_xycoords(shape, distances)
    assert np.array_equal(space.coords(), expected_coords)

    # Test that the extend method returns a new Space object with the correct shape
    extended_shape = tuple(int(s * extend_factor) for s in shape)
    extended_space = space.extend()
    assert extended_space.shape == extended_shape

    # Test that the __post_init__ method calculates the extent correctly
    expected_extent = (-2.5, 2.5, -2.5, 2.5)
    assert np.array_equal(space.extent, expected_extent)

    # Test that asdict method returns a dictionary containing the expected fields
    expected_dict = {
        'shape': shape,
        'distances': distances,
        'space_key': '',
        'extend_factor': 2.0,
        'nifty_domain': RGSpace(shape, distances),
        'center': np.array((0., 0.)),
        'extent': expected_extent
    }
    for key, val in asdict(space).items():
        if isinstance(val, np.ndarray):
            assert np.array_equal(val, expected_dict[key])
        else:
            assert val == expected_dict[key]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


#!/usr/bin/env python3
from dataclasses import dataclass
from typing import Callable, Union, Tuple
from typing_extensions import Self

import nifty8.re as jft


from ..spaces import Space
from ..linear_operators import DeflectionAngle
from .parametric_models import read_parametric_model, build_parametric
from .nonparametric_models import build_nonparametric


MEAN_KEY = 'mean'
PERT_KEY = 'perturbations'
SHEAR_KEY = 'shear'
POISSON_KEY = 'poisson_padding'


@ dataclass
class Model:
    model: jft.Model
    config: dict
    prefix: str
    space: Space

    def parametric(
        self, full_info=False
    ) -> Union[Self, Tuple[Self, jft.Model, jft.Model]]:
        # FIXME: A bit hacky as this assumes that we either have a
        # parametric model, or a hybrid model.
        # I.e. this breaks in case of a nonparametric model

        if self.config.get(MEAN_KEY):
            m, ms, ps = build_parametric(
                self.space.coords(), self.prefix, self.config[MEAN_KEY],
                full_info=True
            )
        else:
            m, ms, ps = build_parametric(
                self.space.coords(), self.prefix, self.config,
                full_info=True
            )
        if full_info:
            return Model(m, self.config, self.prefix, self.space), ms, ps
        return Model(m, self.config, self.prefix, self.space)

    def nonparametric(self) -> Union[Self, Tuple[Self, jft.Model, jft.Model]]:
        if self.config.get(PERT_KEY, None) is None:
            return None
        return Model(
            build_nonparametric(
                self.space, self.prefix, self.config[PERT_KEY]),
            self.config, self.prefix, self.space)

    def deflection(
        self,
        parametric: bool = False,
    ) -> Union[Self, Tuple[Self, jft.Model, jft.Model]]:
        '''Gets the deflection of the model.

        Parameters
        ----------
        parametric : bool
            If True, only the parametric model (and shear) is returned.
        '''

        if parametric and MEAN_KEY in self.config:
            config = self.config[MEAN_KEY]
        else:
            config = self.config

        # All convergence models need to have an analytic deflection
        if check_analytic_deflection(config):
            deflection_conv = build_parametric(
                self.space.coords().reshape(2, -1),
                self.prefix,
                config,
                deflection=True
            )

        else:
            poisson = DeflectionAngle(
                self.space, factor=self.config.get(POISSON_KEY, 2))
            deflection_conv = jft.Model(
                lambda x: poisson(self.model(x)).reshape(2, -1),
                domain=self.domain)

        if self.config.get(SHEAR_KEY):
            shear = build_parametric(
                self.space.coords().reshape(2, -1),
                self.prefix,
                self.config[SHEAR_KEY],
                deflection=True
            )

            ptree = {}
            ptree.update(deflection_conv.domain)
            ptree.update(shear.domain)

            deflection_model = jft.Model(
                lambda x: deflection_conv(x) + shear(x),
                domain=ptree
            )
        else:
            deflection_model = deflection_conv

        return Model(deflection_model, self.config, self.prefix, self.space)

    @property
    def domain(self):
        return self.model.domain

    @property
    def target(self):
        return self.model.target

    def __call__(self, x):
        return self.model(x)


def check_analytic_deflection(convergence_model_config: Model) -> bool:
    if convergence_model_config.get(PERT_KEY, None) is not None:
        return False

    return all(
        hasattr(read_parametric_model(
            model_name.lower().split('_')[0]), 'deflection')
        for model_name in convergence_model_config
    )


BuilderPattern = Callable[[Space, str, dict], Model]


def build_parametric_model(
    space: Space,
    prefix: str,
    model_cfg: dict,
) -> Model:
    '''Build a parametric model according to the provided config.

    Parameters
    ----------
    space : Space
        The space where the model is evaluated
    prefix: str
        Identifier key for the model
    model_cfg: dict
        The config dictionary, the keys in the dictionary are the names of the
        parametric model.
        The values provide the prior distribution for the model paramters.
    '''
    if model_cfg is None:
        return None

    prefix = '_'.join((space.space_key, prefix))
    model = build_parametric(space.coords(), prefix, model_cfg)
    return Model(model, model_cfg, prefix, space)


def build_nonparametric_model(
    space: Space,
    prefix: str,
    model_cfg: dict
) -> Model:
    '''Build a nonparametric model according to the provided config.

    Parameters
    ----------
    space : Space
        The space where the model is evaluated
    prefix: str
        Identifier key for the model
    model_cfg: dict
        The config dictionary, which provides the prior settings of the model.
    '''
    if model_cfg is None:
        return None

    prefix = '_'.join((space.space_key, prefix))
    model = build_nonparametric(space, prefix, model_cfg)
    return Model(model, model_cfg, prefix, space)


# TODO: Change this function to build only a hybrid model.
# It should not decide on its own, this should be taken care of by some other
# function.
def build_hybrid_model(
    space: Space,
    model_key: str,
    model_cfg: dict
) -> Model:
    '''Build a hybrid model according to the provided config.

    Parameters
    ----------
    space : Space
        The space where the model is evaluated
    model_key: str
        Identifier key for the model
    model_cfg: dict
        - 'mean' key provides the prior settings of the mean model.
        - 'perturbations' key provides the prior settings of the nonparametric
            perturbations.
    '''

    if model_cfg is None:
        return None

    prefix = '_'.join((space.space_key, model_key))
    para = build_parametric(space.coords(), prefix, model_cfg[MEAN_KEY])

    if model_cfg.get(PERT_KEY):
        pert = build_nonparametric(space, prefix, model_cfg[PERT_KEY])

        ptree = {}
        for m in [para, pert]:
            ptree.update(m.domain)

        if model_cfg.get('type') in [None, 'multiplication']:
            model = jft.Model(
                lambda x: para(x) * pert(x),
                domain=ptree
            )

        elif model_cfg.get('type') == 'addition':
            model = jft.Model(
                lambda x: para(x) + pert(x),
                domain=ptree
            )

        else:
            raise ValueError(
                f"""Invalid perturbation type '{model_cfg.get('type')}'\n
                (supported: 'multiplication', 'addition')""")

    else:
        model = para

    return Model(model, model_cfg, prefix, space)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from functools import reduce
from typing import Callable, Tuple, Optional, Dict, Union
from dataclasses import dataclass

import numpy as np
from numpy.typing import ArrayLike

import nifty8.re as jft


from .spaces import get_extent, Space
from .space_model import SpaceModel
from .mock_data import create_mock_data
from .utils import load_fits, estimate_snr
from .response import (build_response, load_psf_and_lanczos_shift,
                       build_mask_operator)
from .likelihood import build_gaussian_likelihood
from .nifty_connect import jax_operator


@dataclass
class ImageData:
    data_2d: ArrayLike
    noise_2d: ArrayLike
    mask: ArrayLike
    response: Optional[jft.Model] = None
    response_no_psf: Optional[jft.Model] = None
    pixel_size: float = 1

    def extent(self):
        return get_extent(self.data_2d.shape, self.pixel_size)

    def plot_data(self):
        """Plots source, data, convergence and deflection images."""
        import matplotlib.pyplot as plt

        plot_titles = ['Data', 'Mask', 'Noise']
        plot_data = [
            self.data_2d,
            self.mask,
            self.noise_2d
        ]
        fig, axes = plt.subplots(1, 3)
        for data, title, ax in zip(plot_data, plot_titles, axes.flatten()):
            im = ax.imshow(data, origin='lower')
            fig.colorbar(im, ax=ax)
            ax.set_title(title)
        plt.show()


@dataclass
class MockData:
    source: ArrayLike
    convergence: ArrayLike
    deflection: ArrayLike
    snr: float
    data_2d: ArrayLike
    noise_2d: ArrayLike
    mask: ArrayLike
    response: Optional[jft.Model] = None
    response_no_psf: Optional[jft.Model] = None
    pixel_size: float = 1
    Ls: Optional[ArrayLike] = None
    mock_pos: Optional[Dict[str, ArrayLike]] = None

    def extent(self):
        return get_extent(self.data_2d.shape, self.pixel_size)

    def plot_data(self):
        """Plots source, data, convergence and deflection images."""
        import matplotlib.pyplot as plt

        plot_titles = ['Source', 'Data', 'Convergence', 'Deflection']
        plot_data = [
            self.source,
            self.data_2d,
            self.convergence,
            np.hypot(*self.deflection),
        ]
        fig, axes = plt.subplots(2, 2)
        for data, title, ax in zip(plot_data, plot_titles, axes.flatten()):
            im = ax.imshow(data, origin='lower')
            fig.colorbar(im, ax=ax)
            ax.set_title(f'{title} SnR:{self.snr.mean():.3f}' if
                         title == 'Data' else title)
        plt.show()


def load_mask(data_cfg: dict, data: ArrayLike, noise: ArrayLike) -> ArrayLike:
    """Loads the mask from the configuration file."""
    if data_cfg.get('mask_path', None) is None:
        return np.logical_or(
            np.isnan(data),
            np.isnan(noise))
    else:
        mask = load_fits(data_cfg['mask_path'])
        mask = mask.astype(bool)
        nan_mask = np.logical_or(  # Mask where data or noise is nan
            np.isnan(data),
            np.isnan(noise)
        )
        return np.logical_or(mask, nan_mask)


def load_std(data_cfg: dict, data: ArrayLike) -> Tuple[ArrayLike, float]:
    """Loads the noise from the configuration file."""
    if data_cfg.get('noise_path', None) is None:
        noise_scale = data_cfg['noise_scale']
        return np.full(data.shape, noise_scale)
    else:
        return load_fits(data_cfg['noise_path'])


def load_mock_data(
        cfg_data_mock_config: dict,
        lens_source_model: Tuple[SpaceModel, SpaceModel],
        psf: ArrayLike
) -> MockData:
    """Loads the mock data from the configuration file."""

    lens_plane_model, source_plane_model = lens_source_model

    source, data_2d, convergence, deflection, mock_pos, noise_scale, Ls = \
        create_mock_data(
            lens_plane_model,
            source_plane_model,
            mock_config=cfg_data_mock_config,
            psf=psf,
            seed=cfg_data_mock_config['seed'],
        )

    noise_2d = np.full(data_2d.shape, noise_scale)

    pixel_size = cfg_data_mock_config['data_space']['distance']
    if np.isscalar(pixel_size):
        pixel_size = (float(pixel_size),) * len(data_2d.shape)
    else:
        tmp = np.empty(len(data_2d.shape), dtype=float)
        tmp[:] = pixel_size
        pixel_size = tuple(tmp)

    return MockData(
        data_2d=data_2d,
        noise_2d=noise_2d,
        mask=np.full(data_2d.shape, False),
        source=source,
        convergence=convergence,
        deflection=deflection,
        Ls=Ls,
        snr=estimate_snr(data_2d, noise_2d),
        pixel_size=pixel_size,
        mock_pos=mock_pos,
    )


def load_data(
        cfg_data: dict,
) -> Union[ImageData, MockData]:
    '''Load data from file'''

    data_2d = load_fits(cfg_data['data_path'])
    noise_2d = load_std(cfg_data, data_2d)
    mask = load_mask(cfg_data, data_2d, noise_2d)

    pixel_size = cfg_data['pixel_size']
    if np.isscalar(pixel_size):
        pixel_size = (float(pixel_size),) * len(data_2d.shape)
    else:
        tmp = np.empty(len(data_2d.shape), dtype=float)
        tmp[:] = pixel_size
        pixel_size = tuple(tmp)

    # Some tests
    assert data_2d.shape == noise_2d.shape
    assert data_2d.shape == mask.shape

    # Check if source is provided (Birrer)
    if cfg_data.get('source_path', None) is None:
        return ImageData(
            data_2d=data_2d,
            noise_2d=noise_2d,
            mask=mask,
            pixel_size=pixel_size
        )

    source = load_fits(cfg_data['source_path'])
    return MockData(
        data_2d=data_2d,
        noise_2d=noise_2d,
        mask=mask,
        pixel_size=pixel_size,
        source=source,
        convergence=np.zeros_like(data_2d),
        deflection=(np.zeros_like(data_2d),)*2,
        snr=estimate_snr(data_2d, noise_2d)
    )


DataLoadingStrategy = Callable[
    [dict,  # cfg
     str,  # domain_key
     Space,  # sky_domain
     Optional[Tuple[SpaceModel, SpaceModel]]  # lens_source_model
     ],
    Tuple[jft.Likelihood, Dict[str, Union[ImageData, MockData]]]
]


def MockDataLoadingStrategy(
        cfg: dict,
        sky_key: str,
        sky_domain: Space,
        lens_source_model: Tuple[SpaceModel, SpaceModel],
        nifty_connect: bool = True
) -> Tuple[jft.Likelihood, Dict[str, MockData]]:

    likelihood_key = 'likelihood_01'
    data_dict = {
        likelihood_key: load_mock_data(
            cfg['data']['mock_config'],
            lens_source_model,
            load_psf_and_lanczos_shift(cfg['data']['files'], sky_domain)[0])
    }

    R, response_no_psf = build_response(
        domain_key=sky_key,
        domain=sky_domain,
        data_pixel_size=data_dict[likelihood_key].pixel_size,
        likelihood_key=likelihood_key,
        likelihood_config=cfg['data']['files']
    )

    mask_array = ~data_dict[likelihood_key].mask
    Mask_d = build_mask_operator(R.target, mask_array)

    likelihood = build_gaussian_likelihood(
        data=data_dict[likelihood_key].data_2d[mask_array],
        std=data_dict[likelihood_key].noise_2d[mask_array],
    )

    likelihood = likelihood.amend(Mask_d)
    likelihood = likelihood.amend(R, domain=R.domain)

    data_dict[likelihood_key].response = R
    data_dict[likelihood_key].response_no_psf = response_no_psf
    data_dict[likelihood_key].mask2data = Mask_d

    if nifty_connect:
        from .nifty_connect import build_gaussian_likelihood as bgl
        Mask_d = jax_operator(Mask_d)
        R = jax_operator(R)
        response_no_psf = jax_operator(response_no_psf)
        likelihood = bgl(
            domain=Mask_d.target,
            data=data_dict[likelihood_key].data_2d[mask_array],
            std=data_dict[likelihood_key].noise_2d[mask_array],
        )
        likelihood = (likelihood @ Mask_d) @ R

    return likelihood, data_dict


def RealDataLoadingStrategy(
        cfg: dict,
        sky_key: str,
        sky_domain: Space,
        nifty_connect: bool = True
) -> Tuple[jft.Likelihood, Dict[str, Union[ImageData, MockData]]]:

    data_dict = {}
    likelihoods = []

    for key in cfg['data']['files']:
        data_dict[key] = load_data(cfg['data']['files'][key])
        R, response_no_psf = build_response(
            domain_key=sky_key,
            domain=sky_domain,
            data_pixel_size=data_dict[key].pixel_size,
            likelihood_key=key,
            likelihood_config=cfg['data']['files'][key])

        # Load mask and Mask operator
        Mask_d = build_mask_operator(R.target, ~data_dict[key].mask)

        likelihood = build_gaussian_likelihood(
            data=data_dict[key].data_2d[~(data_dict[key].mask)],
            std=data_dict[key].noise_2d[~(data_dict[key].mask)]
        )
        # likelihood = (likelihood @ Mask_d) @ R
        likelihood = likelihood.amend(Mask_d)
        likelihood = likelihood.amend(R, domain=R.domain)

        data_dict[key].response = R
        data_dict[key].response_no_psf = response_no_psf
        data_dict[key].mask2data = Mask_d

        if nifty_connect:
            from .nifty_connect import build_gaussian_likelihood as bgl
            Mask_d = jax_operator(Mask_d)
            R = jax_operator(R)
            response_no_psf = jax_operator(response_no_psf)
            likelihood = bgl(
                domain=Mask_d.target,
                data=data_dict[key].data_2d[~(data_dict[key].mask)],
                std=data_dict[key].noise_2d[~(data_dict[key].mask)],
            )
            likelihood = (likelihood @ Mask_d) @ R

        likelihoods.append(likelihood)

    likelihood = reduce(lambda x, y: x + y, likelihoods)

    return likelihood, data_dict



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import jax.numpy as jnp

from charm_lensing.spaces import Space, get_klcoords
from numpy import array


class FFTOperator():
    def __init__(self, domain):
        self.domain = domain

    def times(self, field):
        return jnp.fft.fft2(field)

    def inverse(self, field):
        return jnp.fft.ifft2(field)


class ZeroPadder():
    def __init__(self, domain, factor):
        self.domain = domain
        self.target = Space(array(domain.shape)*factor, domain.distances)
        self._zeros = jnp.zeros(self.target.shape)
        self._slice = slice(0, domain.shape[0]), slice(0, domain.shape[1])

    def times(self, field):
        return self._zeros.at[self._slice].set(field)

    def adjoint(self, field):
        return field[self._slice]

    def __repr__(self):
        return 'ZeroPadder: \n domain={} \n target={}'.format(
            self.domain, self.target)


class DeflectionAngle():
    def __init__(self, domain, factor=4):
        self.domain = (domain.shape, domain.distances)
        self.target = ((2,)+domain.shape, domain.distances)

        self.Zp = ZeroPadder(domain, factor)
        self.FFT = FFTOperator(self.Zp.target)
        self.greenskernel = self.deflection_greens_kernel(
            get_klcoords(self.Zp.target.shape, self.Zp.target.distances))

    def __repr__(self):
        return 'domain: {}\ntarget: {}'.format(self.domain, self.target)

    @staticmethod
    def deflection_greens_kernel(kl_coords):
        '''Returns greens-kernel field in Fourier domain'''
        k_square = array(kl_coords[0]**2 + kl_coords[1]**2)
        k_square[0, 0] = 1
        return -1j*kl_coords/(jnp.pi*k_square)

    def deflection_field(self, field):
        tmp = self.FFT.times(self.Zp.times(field))
        return jnp.array((
            self.Zp.adjoint(self.FFT.inverse(self.greenskernel[0]*tmp)).real,
            self.Zp.adjoint(self.FFT.inverse(self.greenskernel[1]*tmp)).real
        ))

    def __call__(self, x):
        return self.deflection_field(x)


class LensingPotential():
    def __init__(self, domain, factor=4):
        self.domain = (domain.shape, domain.distances)
        self.target = (domain.shape, domain.distances)

        self.ZP = ZeroPadder(domain, factor)
        self.FFT = FFTOperator(self.ZP.target)
        self.potential_kernel = self.fourier_potential_kernel(
            get_klcoords(self.ZP.target.shape, self.ZP.target.distances))

    @staticmethod
    def fourier_potential_kernel(kl_coords):
        k_square = array(kl_coords[0]**2 + kl_coords[1]**2)
        k_square[0, 0] = 1.0
        potential_kernel = -1.0 / (k_square*(2.0*jnp.pi**2))
        return potential_kernel

    def potential_field(self, field):
        fft_convolved_field = self.potential_kernel * \
            self.FFT.times(self.ZP.times(field))
        return self.ZP.adjoint(self.FFT.inverse(fft_convolved_field))

    def __call__(self, x):
        return self.potential_field(x)



import matplotlib.pyplot as plt
import numpy as np
from charm_lensing.analysis_tools import chi2, shift_and_resize


def test_chi2():
    data, model = np.array([np.nan, 2, 3]), np.array([1, 2, 3])
    sigma = np.array([5, 2, 3])
    assert chi2(data, model, sigma) == 0

    data, model = np.array([1, 2]), np.array([2, 3])
    sigma = 2.0
    assert chi2(data, model, sigma) == 0.25

    data, model = np.array([1, 2]), np.array([2, 2])
    sigma = np.array([3.0, 2.0])
    assert chi2(data, model, sigma) == 1/9/2


def test_shift():
    # test shift
    i = np.array([[1, 2, 3], [4, 5, 6]])
    r = np.roll(i, (1, 1), axis=(0, 1))
    s, sx, sy = shift_and_resize(i, r, full_info=True)
    assert np.allclose(i, s)
    assert np.allclose(sx, 1)
    assert np.allclose(sy, 1)


def test_resize():
    # test_resize
    print('0.75**2', 0.75**2)
    print('0.25*0.75', 0.25*0.75)
    print('0.25**2', 0.25**2)
    input = np.array([[1, 1, 2, 2],
                      [1, 1, 2, 2],
                      [3, 3, 4, 4],
                      [3, 3, 4, 4]])
    recon = np.array([[1, 2], [3, 4]])
    # o = shift_and_resize(input, recon)
    # assert np.allclose(i, s)

    # import matplotlib.pyplot as plt
    # fig, axes = plt.subplots(1, 2)
    # ax, ay = axes
    # ax.imshow(recon)
    # ay.imshow(o)
    # plt.show()

    assert True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import nifty8 as ift
from typing import Callable, Union


# Constants
SWITCHES = 'switches'
NAME = 'name'
DELTA_E = 'deltaE'
CONV_LVL = 'convergence_level'
ITER_MIN = 'miniter'
ITER_LIM = 'maxiter'
MINIMIZATION = 'minimization'
SAMPLES = 'samples'
N_SAMPLES = 'n_samples'
MODE = 'mode'
KL_MINI = 'kl_mini'

LINEAR_KEYWORDS = ['linear', 'lin', 'linear_sample', 'linear_resample']


def get_config_value(
        key: str, config: dict, index: int, default=1
) -> Union[int, float]:
    '''Returns a configuration value.
    If the value is a list, returns the value at the specified index.'''
    value = config.get(key, default)
    if isinstance(value, list):
        try:
            return value[index]
        except IndexError:
            return value[-1]
    return value


def get_range_index(
        mini_cfg: dict, iteration: int, total_iterations: int) -> int:
    '''Return the index of the switch-range the iteration is in'''

    switches = mini_cfg.get(SWITCHES, [0])
    switches = switches + [total_iterations]

    def inner(iteration: int, ranges: list) -> int:
        """Return the index of the switch-range the iteration is in"""
        for i in range(len(ranges)-1):
            if ranges[i] <= iteration < ranges[i+1]:
                return i

    return inner(iteration, switches)


def controller_factory(
    mini_cfg: dict, total_iterations: int, prefix: str = ''
) -> Callable[[int], ift.AbsDeltaEnergyController]:
    """Creates a Callable which returns the appropriate
    controller for a given configuration and iteration."""

    def get_controller(iteration: int) -> ift.AbsDeltaEnergyController:

        range_index = get_range_index(mini_cfg, iteration, total_iterations)

        delta_name = '_'.join((prefix, DELTA_E))
        conv_lvl_name = '_'.join((prefix, CONV_LVL))
        iter_lim_name = '_'.join((prefix, ITER_LIM))

        name = get_config_value(NAME, mini_cfg, range_index)
        delta_e = get_config_value(delta_name, mini_cfg, range_index)
        conv_level = get_config_value(conv_lvl_name, mini_cfg, range_index)
        iter_limit = get_config_value(iter_lim_name, mini_cfg, range_index)

        return ift.AbsDeltaEnergyController(
            deltaE=delta_e,
            convergence_level=conv_level,
            iteration_limit=iter_limit,
            # name=name
            name=f'{name}: dE={delta_e:.1e}, cl={conv_level}, il={iter_limit}'
        )

    return get_controller


def n_samples_factory(mini_cfg: dict) -> Callable[[int], int]:
    '''Creates a Callable[iterations] which returns the number of samples.'''

    def n_samples(iteration: int) -> int:
        range_index = get_range_index(
            mini_cfg[SAMPLES], iteration, mini_cfg['n_total_iterations'])
        return get_config_value(N_SAMPLES, mini_cfg[SAMPLES], range_index)

    return n_samples


def sample_type_factory(mini_cfg: dict) -> Callable[[int], str]:
    '''Creates a Callable[iterations] which returns the sample type for
    nifty8.re.'''

    def sample_type(iteration: int) -> str:
        range_index = get_range_index(
            mini_cfg[SAMPLES], iteration, mini_cfg['n_total_iterations'])
        return get_config_value(MODE, mini_cfg[SAMPLES], range_index).lower()

    sample_keywords = [
        "linear_sample",
        "linear_resample",
        "nonlinear_sample",
        "nonlinear_resample",
        "nonlinear_update",
    ]
    etxt = "Unknown sample type: {t} at iteration {ii}, known types: {k}"
    for ii in range(mini_cfg['n_total_iterations']):
        t = sample_type(ii)
        assert t in sample_keywords, etxt.format(t=t, ii=ii, k=sample_keywords)

    return sample_type


def linear_sample_kwargs_factory(
    mini_cfg: dict, prefix: str = 'lin'
) -> Callable[[int], dict]:
    """Creates a Callable[iterations] which returns linear sample kwargs for
    nifty8.re."""

    def lin_kwargs_getter(iteration: int) -> dict:
        range_index = get_range_index(
            mini_cfg[SAMPLES], iteration, mini_cfg['n_total_iterations'])

        delta_name = '_'.join((prefix, DELTA_E))
        miniter_name = '_'.join((prefix, ITER_MIN))
        maxiter_name = '_'.join((prefix, ITER_LIM))

        minit = get_config_value(miniter_name, mini_cfg[SAMPLES], range_index)
        maxit = get_config_value(maxiter_name, mini_cfg[SAMPLES], range_index)
        delta_e = get_config_value(delta_name, mini_cfg[SAMPLES], range_index)

        return dict(
            cg_name=f'{prefix}: dE={delta_e}, mi={minit}, il={maxit}',
            cg_kwargs=dict(
                absdelta=delta_e,
                miniter=minit,
                maxiter=maxit
            ))

    return lin_kwargs_getter


def nonlinear_sample_kwargs_factory(
    mini_cfg: dict, prefix: str = 'nonlin'
) -> Callable[[int], dict]:
    '''Creates a Callable[iterations] which returns nonlinear sample kwargs for
    nifty8.re.'''

    def nonlin_kwargs_getter(iteration: int) -> dict:
        range_index = get_range_index(
            mini_cfg[SAMPLES], iteration, mini_cfg['n_total_iterations'])

        delta_name = '_'.join((prefix, DELTA_E))
        miniter_name = '_'.join((prefix, ITER_MIN))
        maxiter_name = '_'.join((prefix, ITER_LIM))
        minit = get_config_value(miniter_name, mini_cfg[SAMPLES], range_index)
        maxit = get_config_value(maxiter_name, mini_cfg[SAMPLES], range_index)
        delta_e = get_config_value(delta_name, mini_cfg[SAMPLES], range_index)

        cg_delta_name = '_'.join((prefix, 'cg', DELTA_E))
        cg_miniter_name = '_'.join((prefix, 'cg',  ITER_MIN))
        cg_maxiter_name = '_'.join((prefix,  'cg', ITER_LIM))
        cg_delta = get_config_value(
            cg_delta_name, mini_cfg[SAMPLES], range_index, default=None)
        cg_minit = get_config_value(
            cg_miniter_name, mini_cfg[SAMPLES], range_index, default=None)
        cg_maxit = get_config_value(
            cg_maxiter_name, mini_cfg[SAMPLES], range_index, default=None)

        nl_name = f'{prefix}: xtol={delta_e}, mi={minit}, il={maxit}'
        cg_name = f'{prefix}_cg: dE={cg_delta}, mi={cg_minit}, il={cg_maxit}'
        return dict(
            minimize_kwargs=dict(
                name=nl_name,
                xtol=delta_e,
                miniter=minit,
                maxiter=maxit,
                cg_kwargs=dict(
                    name=cg_name,
                    absdelta=cg_delta,
                    miniter=cg_minit,
                    maxiter=cg_maxit
                )))

    return nonlin_kwargs_getter


def kl_kwargs_factory(
        mini_cfg: dict, prefix: str = 'kl'
) -> Callable[[int], dict]:
    '''Creates a Callable[iterations] which returns kl minimization kwargs for
    nifyt8.re.'''

    def kl_kwargs_getter(iteration: int) -> dict:
        range_index = get_range_index(
            mini_cfg[KL_MINI], iteration, mini_cfg['n_total_iterations'])

        delta_name = '_'.join((prefix, DELTA_E))
        miniter_name = '_'.join((prefix, ITER_MIN))
        maxiter_name = '_'.join((prefix, ITER_LIM))
        minit = get_config_value(miniter_name, mini_cfg[KL_MINI], range_index)
        maxit = get_config_value(maxiter_name, mini_cfg[KL_MINI], range_index)
        delta_e = get_config_value(delta_name, mini_cfg[KL_MINI], range_index)

        return dict(
            minimize_kwargs=dict(
                name=f'{prefix}: dE={delta_e}, mi={minit}, il={maxit}',
                absdelta=delta_e,
                miniter=minit,
                maxiter=maxit,
                cg_kwargs=dict(name=f'{prefix}CG')
            ))

    return kl_kwargs_getter


def linear_sampling_factory(
        mini_cfg: dict
) -> Callable[[int], ift.AbsDeltaEnergyController]:
    '''Creates a Callable[iterations] which returns the
    LinearSamplingController.'''
    return controller_factory(
        mini_cfg[SAMPLES], mini_cfg['n_total_iterations'], 'lin')


def nonlinear_sampling_factory(
        mini_cfg: dict, minimizer_algorithm=ift.NewtonCG
) -> Callable[[int], Union[None, ift.NewtonCG]]:
    """Creates a Callable[iterations] which returns the minimizer
    with the AbsDeltaEnergyController."""

    def nonlinear_sampling(iteration: int) -> Union[None, ift.NewtonCG]:
        range_index = get_range_index(
            mini_cfg[SAMPLES], iteration, mini_cfg['n_total_iterations'])
        mode = get_config_value(MODE, mini_cfg[SAMPLES], range_index)

        if mode.lower() in LINEAR_KEYWORDS:
            return None
        else:
            controller = controller_factory(
                mini_cfg[SAMPLES], mini_cfg['n_total_iterations'], 'nonlin')
            return minimizer_algorithm(controller(iteration))

    return nonlinear_sampling


def minimizer_factory(
        mini_cfg: dict, minimizer_algorithm=ift.NewtonCG
) -> Callable[[int], ift.NewtonCG]:
    """Creates a Callable[iterations] which returns the minimizer
    with the appropriate AbsDeltaEnergyController."""

    def minimizer(iteration: int) -> ift.NewtonCG:
        controller = controller_factory(
            mini_cfg[KL_MINI], mini_cfg['n_total_iterations'], 'kl')
        return minimizer_algorithm(controller(iteration))

    return minimizer


def domain_transition(new_domain, sample_list):
    print('*'*80)
    print('Switching to full model')
    print('*'*80)

    mean, _ = sample_list.sample_stat()
    mean_dict_old = mean.to_dict()
    mean_new = ift.from_random(new_domain) * 0.1
    mean_dict_new = mean_new.to_dict()
    for key, val in mean_dict_new.items():
        if key in mean_dict_old:
            mean_dict_new[key] = mean_dict_old[key]

    return ift.MultiField.from_dict(mean_dict_new)



import pytest

from charm_lensing.models.parametric_models.parametric_prior import (
    build_prior_operator,
    build_parametric_prior,
    shape_adjust,
)

from charm_lensing.models.parametric_models import read_parametric_model


def test_build_prior_operator():
    build_prior_operator('test_01',
                         {'distribution': 'normal',
                          'mean': 1,
                          'sigma': 2}
                         )
    # TODO: need to assert something here, check normal prior

    build_prior_operator('test_02',
                         {'distribution': 'log_normal',
                          'mean': 1,
                          'sigma': 2}
                         )
    # TODO: need to assert something here, check lognormal prior

    build_prior_operator('test_03',
                         {'distribution': 'uniform',
                          'min': 1,
                          'max': 2}
                         )
    # TODO: need to assert something here, check uniform prior

    build_prior_operator('test_03',
                         {'distribution': None,
                          'mean': 1, }
                         )
    # TODO: need to assert something here, check uniform prior

    build_prior_operator('test_04',
                         {'distribution': 'uniform',
                          'min': 1,
                          'max': 2,
                          'transformation': 'log',
                          })
    # TODO: need to assert something here, check log-uniform prior

    with pytest.raises(NotImplementedError):
        build_prior_operator('test_04',
                             {'distribution': 'lonormal',
                              'min': 1,
                              'max': 2})

    with pytest.raises(KeyError):
        build_prior_operator('test_04',
                             {'distribution': 'log_normal',
                              'min': 1,
                              'max': 2})

    out_05 = build_prior_operator('test_05',
                                  {'distribution': 'log_normal',
                                   'mean': 1,
                                   'sigma': 2},
                                  shape=(1, 2, 3))
    assert out_05({'test_05': 1.}).shape == (1, 2, 3)


# Test build_parametric_prior
def test_build_parametric_prior():
    import jax.numpy as jnp

    nfw = read_parametric_model('nfw')
    nfw_prior = {key: {'distribution': 'log_normal',
                       'mean': 1,
                       'sigma': 2} for key, _ in nfw.prior_keys}

    prior = build_parametric_prior(nfw, 'test_01', nfw_prior)
    # TODO: need to assert something here, check normal prior
    pos = {key: jnp.full(prior.domain[key].shape, 0.)
           for key in prior.domain}
    out_01 = prior(pos)

    with pytest.raises(ValueError):
        build_parametric_prior(nfw, 'test_02', {'b': 1})



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from .psf_operator import instantiate_psf
from .utils import load_fits
from .spaces import Space

from charm_lensing.models.parametric_models import (
    read_parametric_model, build_parametric_prior)

import nifty8.re as jft
import numpy as np
import jax.numpy as jnp

from numpy.typing import ArrayLike
from typing import Tuple, Union, Callable

from jax.scipy.signal import fftconvolve

LANCZOS_WINDOW = 3
ZERO_FLUX_KEY = 'zero_flux'


def lanczos_kernel(
    kernel_size: Tuple[int, int],
    shift: Tuple[int, int],
    lanczos_window: int = 3
) -> ArrayLike:
    """
    Function to generate a Lanczos kernel.

    Parameters:
    ----------
    kernel_size:
        The size of the square kernel.
    shift:
        A tuple with the x and y shift.
    lanczos_window:
        The size of the Lanczos window (parameter "a" in the Lanczos function).

    Returns:
    -------
    kernel: The resulting 2D Lanczos kernel as a numpy array.
    """

    # Calculate the coordinates of the center of the kernel
    center_coord = (kernel_size[0] // 2, kernel_size[1] // 2)

    # Calculate the coordinates of each pixel relative to the center
    xx, yy = np.ogrid[:kernel_size[1], :kernel_size[0]]
    relative_y_coords = yy - center_coord[1] - shift[1]
    relative_x_coords = xx - center_coord[0] - shift[0]

    # Use broadcasting to calculate the distance from each pixel to the center
    x = np.sqrt(relative_x_coords ** 2 + relative_y_coords ** 2)

    # Create the Lanczos kernel
    # Usually, lanczos_window is half of the kernel size
    a = lanczos_window
    kernel = np.sinc(x) * np.sinc(x / a)
    kernel[x > a] = 0  # Apply window function

    return kernel


def load_psf_and_lanczos_shift(
    response_cfg: dict,
    kernel_space: Space
) -> Tuple[Union[ArrayLike, None], Tuple[int, int]]:
    '''Load data and PSF for a given key from configuration file.'''

    psf_path = response_cfg.get('psf_path', None)

    if psf_path is None:
        print('No Psf loaded')
        psf = None
        oversampling = None
    else:
        psf = load_fits(psf_path)
        psf = psf / psf.sum()

        # check if psf is oversampled, FIXME: This should be done better
        if 'oversample' in psf_path.split('/')[-1].split('.')[0]:
            oversampling = int(
                psf_path.split('/')[-1].split('.')[0].split('_')[-1]
            )
        else:
            oversampling = 1

    if response_cfg.get('pixel_shift', None) is None:
        coordinate_shift = np.array([0, 0])
    else:
        shift = np.load(response_cfg['pixel_shift'])
        pix_size = response_cfg['pixel_size']
        coordinate_shift = shift * pix_size

    if not np.logical_or(*np.isclose(coordinate_shift, 0)):
        lan_kern = lanczos_kernel(
            kernel_size=kernel_space.shape,
            shift=np.array((coordinate_shift[1], coordinate_shift[0])
                           ) / kernel_space.distances[0],
            lanczos_window=LANCZOS_WINDOW)
        lan_kern = lan_kern / np.sum(lan_kern)

        if psf is not None:
            psf_kernel = fftconvolve(lan_kern, psf, mode='same')
        else:
            psf_kernel = lan_kern
        psf = psf_kernel

    return psf, oversampling


def build_zero_flux(prefix: str, likelihood_config: dict) -> jft.Model:
    model_cfg = likelihood_config.get(ZERO_FLUX_KEY, None)
    if model_cfg is None:
        return None

    prefix = '_'.join([prefix, ZERO_FLUX_KEY])
    zf = read_parametric_model(ZERO_FLUX_KEY)
    zfp = build_parametric_prior(zf, prefix, model_cfg)
    return jft.Model(
        lambda x: zf(zfp(x)),
        domain=zfp.domain
    )


def build_mask_operator(
    domain: jft.ShapeWithDtype,
    mask: ArrayLike
) -> jft.Model:
    assert domain.shape == mask.shape
    mask = jnp.array(mask, dtype=bool)
    return jft.Model(
        lambda x: x[mask],
        domain=domain
    )


def build_integration_operator(
        domain: Space,
        target_distances: Tuple[float, float],
        oversampling: None | int = None
) -> Callable[[ArrayLike], ArrayLike]:
    assert len(domain.shape) == 2
    assert len(target_distances) == len(domain.distances)
    assert all([np.isclose(d, target_distances[0]) for d in target_distances])
    assert all([np.isclose(d, domain.distances[0]) for d in domain.distances])

    if target_distances[0] == domain.distances[0]:
        return lambda x: x

    # check that integration_value is integer
    downsample = [
        target_distances[ii] / domain.distances[ii]
        for ii in range(len(target_distances))
    ]
    assert all([np.isclose(d, np.round(d)) for d in downsample])

    if oversampling is not None:
        etxt = 'integration_value = {iv} != oversampling = {os}'
        assert np.isclose(downsample[0], oversampling), etxt.format(
            iv=downsample[0], os=oversampling)

    downsample = int(downsample[0])

    assert type(downsample) == int
    new_shape = (domain.shape[0] // downsample, downsample,
                 domain.shape[1] // downsample, downsample)
    reshaping_indices = np.reshape(
        np.arange(np.prod(domain.shape)), new_shape)

    def integration(x):
        y = jnp.take(x, reshaping_indices)
        return jnp.sum(y, axis=(1, 3))

    return integration


def build_response(
    domain_key: str,
    domain: Space,
    data_pixel_size: Tuple[float, float],
    likelihood_key: str,
    likelihood_config: dict | None = None
) -> Tuple[jft.Model, jft.Model]:
    '''Build response operator

    Parameters
    ----------
    domain_key: str
        - name of the domain, needed for interfacing with the sky model.
          (Necessary to enable a single evaluation of the sky model.)
    domain: Space - domain
    data_pixel_size: Tuple[float, float] - pixel size of the data
    likelihood_key: str - name of the likelihood
    likelihood_config: dict
        - psf_path:
            path to the psf, or None if no psf is used.
        - pixel_size (optional):
            pixel size of the data, used for the pixel shift.
        - pixel_shift (optional):
            path to the pixel shift, or None if no pixel shift is used.
            The pixel shift must be provided in units of `pixel_size`.
            It provides the shift of the data with respect to the center of the
            `domain`.

        - zero_flux (optional):
            name of the zero-flux model, or None if no zero-flux is used.
    '''

    # TODO: The check if psf, data and domain fit together can be done here.
    # PSF operator & pixel/coordinate shift
    psf_field, oversampling = load_psf_and_lanczos_shift(
        likelihood_config, domain)
    integrator = build_integration_operator(
        domain,
        data_pixel_size,
        oversampling=oversampling)
    psf = instantiate_psf(psf_field)

    # Standard response
    ptree = {domain_key: jft.ShapeWithDtype(domain.shape, float)}
    def R(x): return integrator(psf(x[domain_key]))
    def R_no_psf(x): return integrator(x[domain_key])

    # Zero-flux operator
    zf = build_zero_flux(likelihood_key, likelihood_config)
    if zf is not None:
        ptree.update(zf.domain)
        def R(x): return integrator(psf(x[domain_key])) + zf(x)
        def R_no_psf(x): return integrator(x[domain_key]) + zf(x)

    return (jft.Model(R, domain=jft.Vector(ptree)),
            jft.Model(R_no_psf, domain=jft.Vector(ptree)))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig, Matteo Guardiani


import nifty8.re as jft
from .spaces import Space
from .interpolation import (
    build_linear_interpolation, build_finufft_interpolation)


def build_projection_operator(
        lens_space: Space, source_space: Space, interpolation='bilinear'
) -> jft.Model:
    # Lens equation / Ray casting
    lens_coords_slice = lens_space.get_mask(
        coordinates=True, as_array=True).reshape(2, -1)
    lens_xycoords = lens_space.coords().reshape(2, -1)

    def lens_equation(alpha): return (
        lens_xycoords - alpha[lens_coords_slice].reshape(2, -1)
    )

    if interpolation == 'bilinear':
        print("Using bilinear interpolation")
        ray_trace = build_linear_interpolation(
            ('source', source_space.extend()),
            ('lens', lens_xycoords.shape))
    elif interpolation == 'finufft':
        print("Using finufft interpolation")
        ray_trace = build_finufft_interpolation(
            ('source', source_space.extend()),
            ('lens', lens_xycoords.shape))
    else:
        raise NotImplementedError

    # The lens projection conserves surface brightness, so we
    # 1) devide the source flux by the volume -> surface brightness
    # 2) multiply the lensed surface brightness by the lens volume -> flux
    flux_source_vol = source_space.nifty_domain.dvol
    flux_lens_vol = lens_space.nifty_domain.dvol
    flux_ratio = flux_lens_vol / flux_source_vol

    ptree = {
        'deflection': jft.ShapeWithDtype(lens_coords_slice.shape),
        'source': jft.ShapeWithDtype(source_space.extend().shape)
    }

    return jft.Model(
        lambda x: ray_trace(
            {'lens': lens_equation(x['deflection']),
             'source': x['source']}
        ).reshape(lens_space.shape) * flux_ratio,
        domain=ptree
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import nifty8.re as jft
from ...spaces import Space
from jax.numpy import exp


CFM_KEYS = ['cfm', 'cf', 'correlated', 'correlated_field']
MATERN_KEYS = ['matern', 'mattern', 'matern_field', 'mkf', 'mf']


def add_fluctuations_matern(
    space: Space,
    perturbations_maker: jft.CorrelatedFieldMaker,
    sub_cfg: dict
):
    # Modify dict for matten
    fluc_cfg = sub_cfg['fluctuations']
    fluc_cfg['non_parametric_kind'] = fluc_cfg.get(
        'non_parametric_kind', 'amplitude')
    fluc_cfg['renormalize_amplitude'] = fluc_cfg.get(
        'renormalize_amplitude', False)
    perturbations_maker.add_fluctuations_matern(
        space.shape,
        space.distances,
        **fluc_cfg)


def add_fluctuations(
    space: Space,
    perturbations_maker: jft.CorrelatedFieldMaker,
    sub_cfg: dict
):
    fluc_cfg = sub_cfg['fluctuations']
    fluc_cfg['non_parametric_kind'] = fluc_cfg.get(
        'non_parametric_kind', 'power')
    perturbations_maker.add_fluctuations(
        space.shape,
        space.distances,
        **fluc_cfg)


def build_nonparametric(
    space: Space, model_key: str, pert_cfg: dict
) -> jft.Model:

    supported_models = {
        name: add_fluctuations_matern if name in MATERN_KEYS else
        add_fluctuations for name in CFM_KEYS + MATERN_KEYS
    }

    for key in pert_cfg:
        key_lower = key.split('_')[0].lower()
        if key_lower not in supported_models:
            raise NotImplementedError(
                f"Invalid perturbation model '{key_lower}'"
                f", supporting {list(supported_models.keys())}")

        sub_cfg = pert_cfg[key]
        prefix = '_'.join((model_key, key))  # LEVEL_HIGHER ??

        perturbations_maker = jft.CorrelatedFieldMaker(f'{prefix}_')
        supported_models[key_lower](space, perturbations_maker, sub_cfg)
        perturbations_maker.set_amplitude_total_offset(
            **sub_cfg['amplitude'])

    perturbations = perturbations_maker.finalize()
    model = jft.Model(
        lambda x: exp(perturbations(x)),
        domain=perturbations.domain
    )

    return model



from setuptools import setup, find_packages

setup(
    name='lenscharm',
    version='0.1',
    author='Matteo Guardiani',
    author_email='matteo.guardiani@gmail.com',
    description='A Charming, Bayesian, Strong Lensing Framework',
    packages=find_packages(include=["charm_lensing", "charm_lensing.*"]),
    # package_data={
    #     'module_name': ['config/rules.yml'],
    # },
    # include_package_data=True,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        'numpy',
        'typing-extensions',
        'jax',
        'jaxlib',
        'nifty8',
        'matplotlib',
        'scipy',
        'astropy',
    ],
    url='https://gitlab.mpcdf.mpg.de/ift/lenscharm'
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from .parametric_prior import build_parametric_prior
from .parametric_model import read_parametric_model

from nifty8.re import Model
from jax.numpy import array

from numpy.typing import ArrayLike


def build_parametric(
    coords: ArrayLike,
    prefix: str,
    model_config: dict,
    deflection: bool = False,
    full_info: bool = False
) -> Model:
    '''Build a parametric model according to the provided config.

    Parameters
    ----------
    coords : jnp.ndarray
        The coordinates at which the model is evaluated
    prefix: str
        Identifier key for the model
    model_config: dict
        The config dictionary, the keys in the dictionary are the names of the
        parametric model.
        The values provide the prior distribution for the model paramters.
    deflection: bool
        Whether to build a deflection model
    full_info: bool
        If true, return the model, the individual models and associated priors
    '''

    models, priors, ptree = [], [], {}
    for model_name, model_cfg in model_config.items():
        parametric_prefix = '_'.join((prefix, model_name))
        parametric_model = read_parametric_model(
            model_name.split('_')[0].lower())
        parametric_prior = build_parametric_prior(
            parametric_model, parametric_prefix, model_cfg)

        models.append(parametric_model)
        priors.append(parametric_prior)
        ptree.update(parametric_prior.domain)

    model = Model(
        lambda x: array(
            [m(p(x), coords) for m, p in zip(models, priors)]
        ).sum(axis=0),
        domain=ptree
    )

    if deflection:
        model = Model(
            lambda x: array(
                [m.deflection(p(x), coords) for m, p in zip(models, priors)]
            ).sum(axis=0),
            domain=ptree
        )

    if full_info:
        return model, models, priors

    return model



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from .nfw import Nfw, Cnfw
from .isothermal import Piemd, Dpie, Pjaffe
from .power_law import Spl, Bpl, Cpl
from .gaussian import Gaussian
from .sersic import Sersic
from .shear import Shear
from .exponential_disk import ExponentialDisk
from .linear_models import Constant, ZeroFlux, LinearShift

from typing import Protocol, Tuple
from numpy.typing import ArrayLike


class ParametricModel(Protocol):
    prior_keys: Tuple[Tuple]

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        ...


def read_parametric_model(model_type: str) -> ParametricModel:
    """Convenience function to translate a string into a ParametricModel"""
    mean_models = {
        'nfw': Nfw(),
        'cnfw': Cnfw(),
        'piemd': Piemd(),
        'pie': Piemd(),
        'dpie': Dpie(),
        'pjaffe': Pjaffe(),
        'spl': Spl(),
        'bpl': Bpl(),
        'cpl': Cpl(),
        'gaussian': Gaussian(),
        'sersic': Sersic(),
        'shear': Shear(),
        'exponentialdisk': ExponentialDisk(),
        'constant': Constant(),
        'zero_flux': ZeroFlux(),
        'linearshift': LinearShift(),
    }
    if model_type in mean_models:
        return mean_models[model_type]
    raise NotImplementedError(
        f"Invalid model type '{model_type}', supporting {mean_models.keys()}")


# TODO: This can be used to get rid of the params call and simplify the
# prior_keys code structure
def params_reader(model: ParametricModel, params: dict) -> Tuple[float]:
    """Reads the parameters from a dictionary"""
    return tuple(params[key] for key, _ in model.prior_keys)
